#include "fs/TmpFS.h"
#include "std/errno.h"
#include "std/string.h"
#include "utils/Dynamic.h"
#include "utils/move.h"

namespace TmpFS
{
    struct File
    {
        char* ptr;
        uint64_t pages;
        uint64_t length;
    };

    struct Directory
    {
        Dynamic<VFS::Node*> files;
    };
}

VFS::Node* tmpfs_root = nullptr;

Dynamic<TmpFS::Directory> tmpfs_dirs;
Dynamic<TmpFS::File> tmpfs_files;

extern uint64_t clock_now();

VFS::Node* TmpFS::get()
{
    if (tmpfs_root) return tmpfs_root;
    tmpfs_root = new VFS::Node;
    tmpfs_root->length = 0;
    tmpfs_root->inode = 0;
    tmpfs_root->type = VFS_DIRECTORY;
    tmpfs_root->find_func = TmpFS::finddir;
    tmpfs_root->readdir_func = TmpFS::readdir;
    tmpfs_root->mkdir_func = TmpFS::mkdir;
    tmpfs_root->mode = 0755;
    tmpfs_root->uid = tmpfs_root->gid = 0;
    tmpfs_root->atime = tmpfs_root->ctime = tmpfs_root->mtime = clock_now();
    strncpy(tmpfs_root->name, "tmpfs", sizeof(tmpfs_root->name));
    tmpfs_dirs.ensure_capacity(16);
    tmpfs_files.ensure_capacity(16);
    tmpfs_dirs.set_expand_rate(16);
    tmpfs_files.set_expand_rate(16);
    ensure(tmpfs_dirs.push({{}}));
    return tmpfs_root;
}

VFS::Node* TmpFS::finddir(VFS::Node* node, const char* filename)
{
    if (!node) return nullptr;
    if (node->inode >= tmpfs_dirs.size()) return nullptr;
    auto& dir = tmpfs_dirs[node->inode];
    for (size_t i = 0; i < dir.files.size(); i++)
    {
        if (!strncmp(dir.files[i]->name, filename, sizeof(dir.files[i]->name))) return dir.files[i];
    }
    return nullptr;
}

VFS::Node* TmpFS::readdir(VFS::Node* node, long offset)
{
    if (!node) return nullptr;
    if (node->inode >= tmpfs_dirs.size()) return nullptr;
    auto& dir = tmpfs_dirs[node->inode];
    if (offset < 0) return nullptr;
    if ((size_t)offset >= dir.files.size()) return nullptr;
    return dir.files[offset];
}

int TmpFS::mkdir(VFS::Node* node, const char* name, mode_t mode)
{
    if (node->inode >= tmpfs_dirs.size()) return -EINVAL;
    if (!(node->type & VFS_DIRECTORY)) return -ENOTDIR;
    auto& parent = tmpfs_dirs[node->inode];
    uint64_t inode = tmpfs_dirs.size();
    VFS::Node* new_node = new VFS::Node;
    new_node->inode = inode;
    new_node->find_func = TmpFS::finddir;
    new_node->readdir_func = TmpFS::readdir;
    new_node->mkdir_func = TmpFS::mkdir;
    new_node->length = 0;
    new_node->type = VFS_DIRECTORY;
    new_node->mode = mode;
    new_node->uid = new_node->gid = 0;
    new_node->atime = new_node->ctime = new_node->mtime = clock_now();
    strncpy(new_node->name, name, sizeof(new_node->name));
    ensure(tmpfs_dirs.push({{}}));
    node->length++;
    ensure(parent.files.push(new_node));
    return 0;
}